/*
 * Copyright (c) 1996-1998 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */


/************************************************************************
 * itembase.c: implementation of Item and category subclasses
************************************************************************/
#include "threadP.h"


/************************************************************************
 * Bindable implementation
************************************************************************/


/************************************************************************
 * Bindable::WriteTo
 *
 * Dumps our state to the given output stream.
************************************************************************/
#if defined(THREADLIB_DEBUG)
void
Bindable::WriteTo( ostream& s ) const
{
	// dump our base class state
	Item::WriteTo( s );

	// dump our state
	s << "tid = " << tid << endl;
}
#endif // defined(THREADLIB_DEBUG)


/************************************************************************
 * IOBindable implementation
************************************************************************/


void
IOBindable::Init( thread_t tid,
					bool isspecial,
					int (*will_block_fcn)(void*),
					void* wbf_arg )
{
	Bindable::Init( tid );
	special = isspecial;
	will_block = will_block_fcn;
	desc = wbf_arg;
}



/************************************************************************
 * IOBindable::IsBufferReady
 *
 * Checks to see if we are a buffered bindable and if we've already 
 * received any messages.
 *
 * returns:
 *   `true' if there are waiting buffered messages.  
 *          `*sizep' is updated if not polling.
 *
 *   `false' if there are no buffered messages.
************************************************************************/

bool
IOBindable::IsBufferReady(unsigned* sizep, unsigned do_poll) const
{
	bool ready = false;

    assert(thr_curr);


	// check if we are a buffered bindable and would block
	if( (will_block != NULL) && !(will_block( desc )) )
	{
		if (!do_poll) {
			*sizep = 0;
		}
		ready = true;
	}
	return ready;
}


void
IOBindable::HandleMessage( void* buf, unsigned* sizep )
{
	unsigned nbytes = 0;


	// dequeue the message iff we're not special
	if( !special )
	{
		if( this->Read( buf, *sizep, &nbytes ) != THR_OKAY )
		{
			thr_report_error( "*** IOBindable::HandleMessage failed ***\n" );
			abort();
		}
	}

	*sizep = (unsigned)nbytes;
}

/************************************************************************
 * IOBindable::WriteTo
 *
 * Dumps our state to the given output stream.
************************************************************************/

#if defined(THREADLIB_DEBUG)
void
IOBindable::WriteTo( ostream& s ) const
{
	// dump our base class state
	Bindable::WriteTo( s );

	// dump our state
	s << "special = " << special << endl;
	s << "will_block = " << will_block << endl;
	s << "desc = " << desc << endl;
}
#endif // defined(THREADLIB_DEBUG)



